# -*- coding: utf-8 -*-
import warnings
from fsqlfly.models import Resource, Namespace
from copy import deepcopy
import yaml
from typing import List, Dict, Optional
from sqlalchemy import create_engine, inspect
from sqlalchemy.dialects.mysql.types import (_StringType, _FloatType, _IntegerType, DECIMAL, DATETIME, TIMESTAMP, BIT,
                                             TIME,
                                             BIGINT)

from sqlalchemy.sql.sqltypes import DATE, _Binary
from fsqlfly.canal_manager.canal_consumer import CanalConsumer


class Category:
    MYSQL = 'mysql'
    KAFKA = 'kafka'
    KAFKA_DELETE_AND_CREATE = 'kafka_dc'
    KAFKA_UPDATE = 'kafka_u'
    ES = 'es'


DEFAULT_FORMAT = {"property-version": 1, "type": "json", "derive-schema": True}


class LoadMySQLResource(CanalConsumer):
    help = 'Load Resource From MySQL'
    kafka_use_row_time = True
    kafka_row_time_name = 'mysql_row_time'
    kafka_process_time_name = 'flink_process_time'
    kafka_row_time_delay_second = 5
    kafka_connector_start_mode = "latest-offset"

    def add_arguments(self, parser):
        parser.add_argument('--host', action='store', help='mysql host')
        parser.add_argument('--database', action='store', help='mysql database')
        parser.add_argument('--category', action='store', help='from system', default="mysql",
                            choices=['mysql', 'es', 'kafka'])
        parser.add_argument('--tables', action='store', help='mysql tables', default="*")
        parser.add_argument('--kafka-bootstrap', action='store', help='kafka bootstrap servers',
                            default="localhost:9092", dest='kafka')
        parser.add_argument('--es-hosts', action='store', help='elastic search host port,sepe by ,',
                            default="http://localhost:9200", dest='es')
        parser.add_argument('--port', action='store', help='mysql port', default=3306, type=int)
        parser.add_argument('--password', action='store', help='mysql password, if not set must input when running',
                            default=None)
        parser.add_argument('--username', action='store', help='mysql username', default='root')
        parser.add_argument('--kafka_use_row_time', action='store_true', help='kafka use row time from mysql',
                            default=True)
        parser.add_argument('--kafka_row_time_delay_second', action='store', type=int,
                            help='kafka row time delay second',
                            default=None)

        parser.add_argument('--kafka_connector_start_mode', action='store', type=int,
                            help='kafka connector start mode support [earliest-offset|latest-offset|group-offsets|specific-offsets]',
                            default=None)
        parser.add_argument('--kafka_process_time_name', action='store', type=int,
                            help='kafka process time name',
                            default=None)
        self.add_arguments_for_canal_suffix(parser)

    def _init_global_args(self, options: dict):
        es_hosts = []
        for url in options['es'].split(','):
            protocol, h_p = url.split('//')
            host, port = h_p.split(':')
            es_hosts.append(dict(hostname=host, port=int(port), protocol=protocol))

        if len(es_hosts) < 1:
            warnings.warn("es host not good %s" % options['es-hosts'])

        self.ES_HOSTS = es_hosts
        self.JDBC_USER = options['username']
        self.JDBC_PASSWORD = options['password']

        host, port, database = options['host'], options['port'], options['database']
        self.JDBC_URL = f"jdbc:mysql://{host}:{port}/{database}?useSSL=false"
        self.NAMESPACE = database

        self.KAFKA_BOOTSTRAP_SERVERS = options['kafka']

    def _gen_kafka_connector(self, topic_name: str, typ: str, category: str) -> dict:
        return {
            "type": "kafka",
            "property-version": 1,
            "version": "universal",
            "topic": topic_name,
            "startup-mode": self.kafka_connector_start_mode,
            "properties": [
                {
                    "key": "bootstrap.servers",
                    "value": self.KAFKA_BOOTSTRAP_SERVERS,
                },
                {
                    "key": "group.id",
                    "value": f"{topic_name}_{typ}_{category}",
                }
            ]
        }

    def _gen_mysql_connector(self, table_name: str) -> dict:
        return {
            "type": "jdbc",
            "url": self.JDBC_URL,
            "table": table_name,
            "driver": "com.mysql.jdbc.Driver",
            "username": self.JDBC_USER,
            "password": self.JDBC_PASSWORD,

        }

    def _gen_es_connector(self, index: str) -> dict:
        return {
            "type": "elasticsearch",
            "version": "6",
            "hosts": self.ES_HOSTS,
            "index": index,
            "document-type": "data",
        }

    @classmethod
    def _gen_name(cls, name: str, typ: str, category: str) -> str:
        return name + '_' + category

    @property
    def canal_execute_time_name_field(self) -> dict:
        if self.kafka_use_row_time:
            row_time = {
                "timestamps": {
                    "type": "from-field",
                    "from": self.canal_execute_time_name
                },
                "watermarks": {
                    "type": "periodic-bounded",
                    "delay": self.kafka_row_time_delay_second * 1000
                }
            }
            return dict(name=self.kafka_row_time_name, type='TIMESTAMP', rowtime=row_time)
        return dict(name=self.canal_execute_time_name, type='BIGINT')

    @property
    def canal_event_type_name_field(self) -> dict:
        return dict(name=self.canal_event_type_name, type='INT')

    @property
    def kafka_process_time_name_field(self) -> dict:
        return dict(name=self.kafka_process_time_name, type='TIMESTAMP', proctime=True)

    @property
    def canal_kafka_added_schema(self):
        return self.canal_execute_time_name + " TIMESTAMP, " + self.canal_event_type_name + " INT"

    BOOLEAN_TYPE = "INT"

    def _generate_json_format(self, fields: list, category: str) -> dict:
        def _get_schema(fs: list) -> str:
            return 'ROW<{}>'.format(",".join(fs))

        data = {"property-version": 1, "type": "json"}

        if category == Category.KAFKA_DELETE_AND_CREATE:
            fs = ["{} {}".format(x['name'], x['type']) for x in fields] + [self.canal_kafka_added_schema]
            data["schema"] = _get_schema(fs)
        elif category == Category.KAFKA_UPDATE:
            fs = []
            fmt = "{name} {type}"
            for field in fields:
                k, v = field['name'], field['type']
                fs.append(fmt.format(name=k + self.canal_before_column_suffix, type=v))
                fs.append(fmt.format(name=k + self.canal_after_column_suffix, type=v))
                fs.append(fmt.format(name=k + self.canal_update_suffix, type=self.BOOLEAN_TYPE))
            fs.append(self.canal_kafka_added_schema)
            data['schema'] = _get_schema(fs)
        else:
            data["derive-schema"] = True

        return data

    def _gen_schema(self, fields: List[Dict[str, str]], category: str) -> list:
        if category == Category.KAFKA_DELETE_AND_CREATE:
            fields.append(self.canal_execute_time_name_field)
            fields.append(self.canal_event_type_name_field)
            fields.append(self.kafka_process_time_name_field)
            return fields
        elif category == Category.KAFKA_UPDATE:
            rlt = []
            for field in fields:
                k, v = field['name'], field['type']
                rlt.append(dict(name=k + self.canal_before_column_suffix, type=v))
                rlt.append(dict(name=k + self.canal_after_column_suffix, type=v))
                rlt.append(dict(name=k + self.canal_update_suffix, type=self.BOOLEAN_TYPE))
            rlt.append(self.canal_execute_time_name_field)
            rlt.append(self.kafka_process_time_name_field)
            return rlt
        else:
            return fields

    def _build_yml(self, name: str, typ: str, category: str, columns: List[dict]) -> str:
        data = dict()
        fields = []
        for col in columns:

            # print(col['type'], type(col['type']))
            _t = col['type']
            _n = col['name']
            _c = col['comment']
            t_name = None
            # print(str(_t))

            if isinstance(_t, _StringType) or isinstance(_t, _Binary):
                t_name = "STRING"
            elif isinstance(_t, BIT):
                t_name = "INT"
            elif isinstance(_t, _IntegerType):
                t_name = "INT"
                if isinstance(_t, BIGINT):
                    t_name = "BIGINT"
            elif isinstance(_t, _FloatType):
                t_name = "DOUBLE"
            elif isinstance(_t, DECIMAL):
                t_name = "DECIMAL"
            elif isinstance(_t, DATETIME):
                t_name = "TIMESTAMP"
            elif isinstance(_t, TIMESTAMP):
                t_name = "TIMESTAMP"
            elif isinstance(_t, DATE):
                t_name = "DATE"
            elif isinstance(_t, TIME):
                t_name = "TIME"

            assert t_name is not None
            # fields.append(dict(name=_n, type=t_name))
            fields.append(dict(name=_n, type=t_name))

        if category == Category.MYSQL:
            data['connector'] = self._gen_mysql_connector(name)
        elif category.startswith(Category.KAFKA):
            data['connector'] = self._gen_kafka_connector(f"{self.NAMESPACE}_{name}_{category}", typ, category)
            data['update-mode'] = 'append'
            data['format'] = self._generate_json_format(fields, category)
        elif category == Category.ES:
            data['connector'] = self._gen_es_connector(f"{self.NAMESPACE}_{name}")
            data['format'] = DEFAULT_FORMAT
            data['update-mode'] = 'append'
        else:
            raise Exception("Not Support category")
        data['schema'] = self._gen_schema(fields, category)
        data['type'] = typ

        rlt = yaml.dump(data)
        return rlt

    def create_resource_data(self, name: str, typ: str, category: str, columns: List[dict],
                             table_comment: Optional[str]) -> dict:
        print('create ', name, ' :', typ, '--', category, '@@', table_comment)
        obj = dict()
        obj['yaml'] = self._build_yml(name, typ, category, columns)
        obj['name'] = self._gen_name(name, typ, category)
        obj['info'] = table_comment
        obj['typ'] = typ
        return obj

    def create_resource(self, n: str, typ: str, category: str, resource: dict, columns: List[dict],
                        table_comment: Optional[str],
                        namespace_id: int):
        data = self.create_resource_data(n, typ, category, columns, table_comment)
        name = data['name']
        print('name is ', name, end='')
        data['namespace_id'] = namespace_id
        data['is_publish'] = True
        data['is_available'] = True
        if name not in resource:
            res = Resource.create(**data)
            print(' create a resource', res.id, res.name)
        else:
            resource = Resource.select().where(Resource.id == resource[name]).first()
            for k, v in data.items():
                setattr(resource, k, v)
            resource.save()
            print('update a resource ', resource.id, resource.name)

    def handle(self, *args, **options):
        if options['password'] is None:
            warnings.warn('you not set mysql password please input in command line!!!')
            password = input("Your MySQL Password:")
            options['password'] = password
        else:
            password = options['password']

        self._init_global_args(options)

        category = options['category']
        full_namespace = self.NAMESPACE + '_' + category

        namespace = Namespace.select().where(Namespace.name == full_namespace).first()

        if namespace is None:
            namespace = Namespace.create(name=full_namespace, info="auto generate by script")
        username, host, port, database = options['username'], options['host'], options['port'], options['database']
        db_url = f"mysql+pymysql://{username}:{password}@{host}:{port}/{database}?charset=utf8"
        namespace_id = namespace.id

        engine = create_engine(db_url)
        insp = inspect(engine)
        tables_str = options['tables']
        table_names = set(insp.get_table_names())
        if tables_str == '*':
            need_tables = set(table_names)
        else:
            need_tables = set(tables_str.strip().split(','))

        resources = {x.name: x.id for x in Resource.select().where(Resource.namespace == namespace).objects()}
        for n in table_names:
            if n not in need_tables:
                continue
            t_comment = insp.get_table_comment(n)['text']
            cs = insp.get_columns(n)
            # self.create_resource(n, 'sink-table', category, resources, cs, t_comment, namespace_id)
            if category != Category.ES:
                if category == Category.KAFKA:
                    self.create_resource(n, 'source-table', Category.KAFKA_DELETE_AND_CREATE, resources, cs, t_comment,
                                         namespace_id)
                    self.create_resource(n, 'source-table', Category.KAFKA_UPDATE, resources, cs, t_comment,
                                         namespace_id)
                    self.create_resource(n, 'both', category, resources, cs, t_comment, namespace_id)
                else:
                    self.create_resource(n, 'both', category, resources, cs, t_comment, namespace_id)
            else:
                self.create_resource(n, 'sink-table', category, resources, cs, t_comment, namespace_id)

        table_nums = len(need_tables)
        print(f'Successfully load {table_nums} table from {database}: {tables_str}')



# -*- coding:utf-8 -*-
import os
import logging
import tornado.ioloop
import tornado.web
import tornado.log
from tornado.options import parse_command_line, options
from terminado.management import NamedTermManager
from logzero import logger
from fsqlfly import settings
from fsqlfly import handles


class IndexHandler(tornado.web.RequestHandler):
    def get(self, *args, **kwargs):
        self.set_header('content-type', 'text/html')
        self.write(settings.INDEX_HTML)


def run_web():
    application = tornado.web.Application(
        handles.all_handles + [
            (r"/static/(.*)", tornado.web.StaticFileHandler,
             {"path": settings.FSQLFLY_STATIC_ROOT}),
            (r"/(.*)", IndexHandler),
        ],
        cookie_secret=settings.FSQLFLY_COOKIE_SECRET,
        login_url='/login',
        debug=settings.FSQLFLY_DEBUG,
        autoreload=settings.FSQLFLY_DEBUG,
        terminal_manager=settings.TERMINAL_MANAGER,
    )
    logger.info("start running on http://localhost:{} ... ".format(settings.FSQLFLY_WEB_PORT))
    application.listen(settings.FSQLFLY_WEB_PORT)
    # tornado.options.options.logging = 'error'
    tornado.options.options.logging = None
    tornado.options.options.log_to_stderr = False
    # parse_command_line()


    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    run_web()



import time
import logzero
from logzero import logger
from functools import partial
from collections import defaultdict
from typing import Callable, Optional
from fsqlfly.models import Transform
from requests import Session
from datetime import datetime, date
from fsqlfly.workflow import run_transform


def _set_log_file(file: str):
    logzero.logfile(file, maxBytes=int(1e6), backupCount=3)


class FlinkJobDaemon:

    def __init__(self,
                 flink_host: str,
                 frequency: int,
                 max_try: int,
                 login_file: str,
                 max_req_try: int = 5,
                 stop_handle: Optional[Callable] = None):
        _set_log_file(login_file)
        self.flink_host = flink_host
        self.frequency = frequency
        self.max_try = max_try
        self.stop_handle = stop_handle
        self.max_req_try = max_req_try
        self.session = Session()

    def request(self, func: Callable, try_times: int = 0):
        try:
            return func()
        except Exception as err:
            if try_times > self.max_req_try:
                print('Meet Exception please check FSQLFLY_FINK_HOST is correct')
                raise err
            else:
                return self.request(func, try_times + 1)

    def run(self):
        host = self.flink_host
        sleep = self.frequency
        max_print_times, print_times = max(60 / sleep, 1), 0
        session = Session()
        logger.info('Start Running Flink Job Damon')
        all_jobs = dict()
        run_times = defaultdict(lambda: defaultdict(int))
        while True:
            today = str(date.today())
            start_time = time.time()
            query = (Transform.is_publish == True, Transform.is_deleted == False, Transform.is_available == True)
            job_names = {"{}_{}".format(x.id, x.name): x for x in
                         Transform.select().where(*query).order_by(Transform.id.desc()).objects()}

            jobs_url = host + '/jobs'
            js = session.get(jobs_url).json()
            for x in js['jobs']:
                j_id = x['id']
                if j_id not in all_jobs:
                    all_jobs[j_id] = session.get(host + '/jobs/' + j_id).json()
                else:
                    all_jobs[j_id]['state'] = x['status']

            run_jobs = set()
            for k, v in all_jobs.items():
                full_name = v['name']
                for name, obj in job_names.items():
                    if full_name.startswith(name):
                        if v['state'] == 'RUNNING':
                            run_jobs.add(name)

            for k, v in job_names.items():
                if k not in run_jobs:
                    if run_times[today][k] > self.max_try:
                        logger.error('job run too many times one day {}'.format(k))
                    else:
                        run_times[today][k] += 1
                        logger.info('job {} begin run '.format(k))
                        is_ok, r = run_transform(v)
                        if not is_ok:
                            logger.error(r)
                        if self.stop_handle:
                            self.stop_handle(k)

            end_time = time.time()

            cost = end_time - start_time

            if print_times > max_print_times:
                logger.info(" ".join([str(datetime.now())[:19], ' damon cost ', '%.2f' % cost, ' second', ' will sleep ', str(int(sleep - cost))]))
                print_times = 0
            else:
                print_times += 1

            if cost < sleep:
                time.sleep(sleep - cost)



import unittest
from fsqlfly.workflow import run_debug_transform
from terminado import NamedTermManager
from fsqlfly.tests.base_test import BaseTestCase


class MyTestCase(BaseTestCase):
    def test_something(self):
        run_debug_transform(dict(), NamedTermManager(shell_command=['ls']))


if __name__ == '__main__':
    unittest.main()



import logging
import datetime
from playhouse.shortcuts import model_to_dict
from peewee import *
from fsqlfly.settings import DATABASE
from fsqlfly.utils.strings import dict2underline, dict2camel


class BaseModel(Model):
    is_deleted = BooleanField(default=False)
    is_publish = BooleanField(default=False)
    is_available = BooleanField(default=True)
    create_at = DateTimeField(default=datetime.datetime.now)
    update_at = DateTimeField(default=datetime.datetime.now)

    class Meta:
        database = DATABASE

    def save(self, *args, **kwargs):
        self.update_at = datetime.datetime.now()
        return super(BaseModel, self).save(*args, **kwargs)

    @classmethod
    def _rename_foreign(cls, key: str) -> str:
        return key + '_id' if key in ('namespace', 'resource') else key

    @classmethod
    def _get_foreign_id(cls):
        pass

    def to_dict(self, to_camel=False) -> dict:
        _rename = self._rename_foreign
        data = {_rename(k): v for k, v in model_to_dict(self, recurse=False).items()}
        return dict2camel(data) if to_camel else data


class Namespace(BaseModel):
    name = CharField(max_length=256, unique=True)
    info = CharField(max_length=2048, null=True)
    avatar = CharField(max_length=2048, null=True)

    class Meta:
        table_name = "namespace"


class FileResource(BaseModel):
    name = CharField(max_length=128, unique=True)
    info = CharField(max_length=1024, null=True)
    real_path = CharField(max_length=2048)

    class Meta:
        table_name = "file_resource"


class Resource(BaseModel):
    name = CharField(max_length=256, unique=True)
    info = CharField(max_length=2048, null=True)
    typ = CharField(max_length=16)
    yaml = TextField(default='')
    namespace = ForeignKeyField(Namespace, on_delete='NO ACTION', null=True)

    class Meta:
        table_name = "resource"


class Functions(BaseModel):
    name = CharField(max_length=256, unique=True)
    function_from = CharField(max_length=16, default="class")
    class_name = CharField(max_length=256)
    constructor_config = CharField(max_length=2048)
    resource = ForeignKeyField(FileResource, on_delete='NO ACTION')

    class Meta:
        table_name = "functions"


class Transform(BaseModel):
    name = CharField(max_length=256, unique=True)
    info = CharField(max_length=2048, null=True)
    sql = TextField(default='')
    require = TextField(default='')
    yaml = TextField(default='')
    namespace = ForeignKeyField(Namespace, on_delete='NO ACTION', null=True)

    class Meta:
        table_name = "transform"


TABLES = [Namespace, Resource, FileResource, Resource, Transform, Functions]


def delete_all_tables(force: bool = False):
    if not force:
        word = input('Are you delete all tables (Y/n)')
        if word.strip().upper() != 'Y':
            return
    logging.info("begin delete all tables")
    DATABASE.drop_tables(TABLES)


def create_all_tables():
    DATABASE.create_tables(TABLES)



# -*- coding:utf-8 -*-
import os
import sys
import argparse
import logzero
from os.path import join, dirname, abspath


def run_webserver(commands: list):
    from fsqlfly.app import run_web
    try:
        run_web()
    except KeyboardInterrupt:
        logzero.logger.info("Stop Web...")


def run_canal(commands: list):
    from fsqlfly.canal_manager.canal_consumer import CanalConsumer
    CanalConsumer().execute()


def run_load_mysql_resource(commands: list):
    from fsqlfly.canal_manager.load_mysql_resource import LoadMySQLResource
    LoadMySQLResource().execute()


def run_job_daemon(commands: list):
    from fsqlfly import settings
    from fsqlfly.job_manager.daemon import FlinkJobDaemon
    daemon = FlinkJobDaemon(settings.FSQLFLY_FINK_HOST,
                            settings.FSQLFLY_JOB_DAEMON_FREQUENCY,
                            settings.FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY,
                            settings.FSQLFLY_JOB_LOG_FILE)
    print('daemon running...')
    daemon.run()


def init_db(commands: list):
    from fsqlfly.models import create_all_tables

    create_all_tables()


def reset_db(commands: list):
    from fsqlfly.models import delete_all_tables
    conformed_parser = argparse.ArgumentParser("Conformed")
    conformed_parser.add_argument('-f', '--force', type=bool, default=False, help='force running')
    args = conformed_parser.parse_args(commands)
    delete_all_tables(force=args.force)


def run_echo_env(commands: list):
    if len(commands) > 0:
        out = open(commands[0], 'w')
    else:
        out = sys.stdout

    lib_path = dirname(abspath(__file__))
    print(open(join(lib_path, 'env.template'), 'r').read(), file=out)


def main():
    support_command = {
        "echoenv": run_echo_env,
        "webserver": run_webserver,
        "initdb": init_db,
        "resetdb": reset_db,
        "jobdaemon": run_job_daemon,
        "loadmysql": run_load_mysql_resource,
        "canal": run_canal,
    }
    args = sys.argv[1:]
    method = args[0] if len(args) > 0 else 'help'
    if len(args) == 0 or method in ('-h', 'help') or method not in support_command:
        print("Usage : fsqlfly [-h] {}".format('|'.join(list(support_command.keys()) + ['help'])))
        return
    process = support_command[args[0]]
    process(sys.argv[2:])


if __name__ == '__main__':
    main()



# -*- coding:utf-8 -*-
import time
import re
import traceback
import math
from io import StringIO
from typing import List, Any, Dict
from datetime import datetime
from requests import Session
from logzero import logger
from tornado.web import authenticated
from collections import namedtuple, defaultdict
from fsqlfly.settings import FSQLFLY_FINK_HOST, TEMP_TERMINAL_HEAD
from fsqlfly.models import Transform
from fsqlfly.workflow import run_transform
from fsqlfly.base_handle import BaseHandler
from fsqlfly.utils.response import create_response

JobStatus = namedtuple('JobStatus', ['name', 'job_id', 'status', 'full_name',
                                     'start_time', 'end_time', 'duration'])

FAIL_HEADER = 'FAIL:'
SUCCESS_HEADER = 'SUCCESS:'


class Cache:
    def __init__(self, max_store_seconds=5):
        self.time = dict()
        self.store = dict()
        self._max = max_store_seconds

    def get(self, name: str) -> Any:
        if name not in self.store:
            return None
        assert name in self.time
        if (time.time() - self.time[name]) > self._max:
            del self.time[name]
            del self.store[name]
            return None
        return self.store[name]

    def set(self, name: str, value: Any):
        self.store[name] = value
        self.time[name] = time.time()

    def remove(self, name: str):
        if name in self.time:
            del self.time[name]
        if name in self.store[name]:
            del self.store[name]


class JobControl:
    restart = 'restart'
    stop = 'stop'
    start = 'start'
    cancel = 'cancel'
    RUN_STATUS = 'RUNNING'

    def __contains__(self, item):
        if hasattr(self, item):
            return True
        return False

    def __init__(self, flink_host=None):
        print(flink_host)
        self.host = flink_host
        self.session = Session()
        self.cache = Cache()

    def _get_job_status(self, job_id: str) -> JobStatus:
        def p_time(t: int) -> str:
            return str(datetime.fromtimestamp(t / 1000))[:19] if t > 0 else '-'

        def p_duration(t: int) -> str:
            second = 1000
            minute = second * 60
            hour = minute * 60
            day = hour * 24
            out = list()
            for a, b in [(day, '天'), (hour, '小时'), (minute, '分')]:
                cost = 0
                while t > a:
                    cost += 1
                    t -= a
                if cost > 0:
                    out.append(f'{cost}{b}')
            out.append('{}秒'.format(math.ceil(t / second)))
            return ''.join(out)

        status = self.session.get(self.host + '/jobs/' + job_id).json()
        name = status['name'].split('.', maxsplit=1)[0]
        job_status = JobStatus(name, job_id, status['state'], full_name=status['name'],
                               start_time=p_time(status["start-time"]), end_time=p_time(status["end-time"]),
                               duration=p_duration(status['duration']))
        return job_status

    @property
    def job_status(self) -> List[JobStatus]:
        jobs_url = self.host + '/jobs'
        js = self.session.get(jobs_url).json()
        all_jobs = list()
        for x in js['jobs']:
            j_id = x['id']
            if self.cache.get(j_id) is None:
                status = self._get_job_status(j_id)
                self.cache.set(j_id, status)
            else:
                status = self.cache.get(j_id)

            all_jobs.append(status)
        return all_jobs

    @classmethod
    def get_job_header(cls, transform: Transform, **kwargs) -> str:
        return "{}_{}.{}-".format(transform.id, transform.name, '_' + kwargs['pt'] if 'pt' in kwargs else '')

    def handle_restart(self, transform: Transform, **kwargs) -> str:
        msgs = []
        msgs.append(self.handle_stop(transform))
        msgs.append(self.start_flink_job(transform, **kwargs))
        return SUCCESS_HEADER + '\n'.join(msgs)

    def handle_stop(self, transform: Transform, **kwargs) -> str:
        msgs = []
        header = self.get_job_header(transform, **kwargs)
        kill_jobs = []
        job_status = self.job_status
        for job in job_status:
            if job.status == self.RUN_STATUS and job.name == header:
                logger.debug('add a {} to kill '.format(job.name))
                kill_jobs.append(job.job_id)

        msgs.append('kill {} jobs: {}'.format(len(kill_jobs), ', '.join(str(x) for x in kill_jobs)))

        self.stop_flink_jobs(kill_jobs)
        return SUCCESS_HEADER + '\n'.join(msgs)

    def handle_start(self, transform: Transform, **kwargs) -> str:
        return self.start_flink_job(transform, **kwargs)

    def handle_cancel(self, jid: str, **kwargs):
        self.stop_flink_jobs([jid])
        return 'kill {} '.format(jid)

    def stop_flink_jobs(self, job_ids: List):
        for j_id in job_ids:
            logger.debug('begin stop flink job {}'.format(j_id))
            logger.debug(self.host)
            res = self.session.patch(self.host + '/jobs/' + j_id + '?mode=cancel')
            self.cache.remove(j_id)
            print(res.text)

    @classmethod
    def start_flink_job(cls, transform: Transform, **kwargs) -> str:
        is_ok, txt = run_transform(transform, **kwargs)
        return '{} JOB {}\n{}'.format(SUCCESS_HEADER if is_ok else FAIL_HEADER, transform.name, '' if is_ok else txt)


JobControlHandle = JobControl(FSQLFLY_FINK_HOST)


class JobHandler(BaseHandler):
    @authenticated
    def get(self, mode: str, pk: str):
        handle_name = 'handle_' + mode
        if mode in JobControlHandle and handle_name in JobControlHandle:
            if pk.isdigit():
                transform = Transform.select().where(Transform.id == int(pk)).first()
                if transform is None:
                    return self.write_json(create_response(code=500, msg='job id {} not found!!!'.format(pk)))
            else:
                transform = pk

            data = self.json_body
            try:
                run_res = getattr(JobControlHandle, handle_name)(transform, **data)
            except Exception as e:
                out = StringIO()
                traceback.print_exc(file=out)
                out.seek(0)
                return create_response(msg=out.read(), code=500)
            return self.write_json(create_response(msg=run_res, code=500 if run_res.startswith(FAIL_HEADER) else 0))
        else:
            return self.write_json(create_response(code=500, msg=' {} not support!!!'.format(mode)))

    post = get


def get_latest_transform(refresh_seconds: int = 5) -> Dict[str, dict]:
    def _get():
        job_infos = defaultdict(dict)
        for x in Transform.select().order_by(Transform.id.desc()).objects():
            name = "{}_{}".format(x.id, x.name)
            job_infos[name] = x.to_dict()
        return job_infos

    __name = '__transforms'
    if not hasattr(get_latest_transform, __name):
        setattr(get_latest_transform, __name, (_get(), time.time()))
    objects, t = getattr(get_latest_transform, __name)
    if time.time() - t > refresh_seconds:
        objects = _get()
        setattr(get_latest_transform, __name, (objects, time.time()))
    return objects


class JobList(BaseHandler):
    job_pattern = re.compile(r'\d+_')
    @classmethod
    def is_real_job(cls, name) -> bool:
        return cls.job_pattern.search(name) is not None

    @authenticated
    def get(self):
        target = self.request.arguments.get('id')
        id_filter = target[0].decode() if target else None

        job_infos = get_latest_transform()
        all_jobs = list()
        for job in JobControlHandle.job_status:
            if not self.is_real_job(job.name):
                continue
            if id_filter is not None and not job.name.startswith(id_filter + '_'):
                continue
            base = dict(**job._asdict())
            if job.name in job_infos:
                base.update(job_infos[job.name])
                base['t_id'] = job_infos[job.name]['id']
            elif job.name.startswith(TEMP_TERMINAL_HEAD):
                base['name'] = 'TEMPORARY'
                base['url'] = '/terminal/{}'.format(job.name[len(TEMP_TERMINAL_HEAD):])

            base['id'] = job.job_id
            base['detail_url'] = FSQLFLY_FINK_HOST + '/#/job/{}/overview'.format(job.job_id)

            all_jobs.append(base)

        return self.write_json(create_response(data=all_jobs))


default_handlers = [
    (r'/api/job/(?P<mode>\w+)/(?P<pk>[a-zA-Z0-9]+)', JobHandler),
    (r'/api/job', JobList),
]



import unittest
import json
import importlib
from fsqlfly import settings
from peewee import SqliteDatabase, IntegrityError
import fsqlfly.models
from fsqlfly.tests.base_test import BaseTestCase


class MyTestCase(BaseTestCase):
    def test_namespace_unique(self):
        name = 'abcdd'
        sample = fsqlfly.models.Namespace.create(name=name)
        with self.assertRaises(IntegrityError):
            fsqlfly.models.Namespace.create(name=name)

    def test_model_to_dict(self):
        sample = fsqlfly.models.Namespace.create(name='name')
        self.assertTrue(isinstance(sample.to_dict(), dict))
        self.assertTrue(isinstance(json.dumps(sample.to_dict()), str))


if __name__ == '__main__':
    unittest.main()



from . import api_handler
from . import terminal_handler
from . import crud_handler
from . import file_handler
from . import job_handler

all_handles = (api_handler.default_handlers + file_handler.default_handlers + job_handler.default_handlers +
               terminal_handler.default_handlers + crud_handler.default_handlers)



# -*- coding: utf-8 -*-
import time
import os
import logging
import sys
import json
from logzero import logger
from typing import Optional
from collections import defaultdict
from datetime import datetime
from random import randint
from argparse import ArgumentParser
import fsqlfly.settings

PRINT_EACH = 100
WAIT_TIMES = 0.5


class BaseCommand:
    def execute(self):
        parser = ArgumentParser()
        self.add_arguments(parser)
        print(sys.argv[2:])
        options = parser.parse_args(sys.argv[2:])
        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop('args', ())
        output = self.handle(*args, **cmd_options)
        return output

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        pass


class CanalConsumer(BaseCommand):
    help = 'Canal Consumer'
    bootstrap_servers = "localhost:9092"
    canal_execute_time_name = 'MYSQL_DB_EXECUTE_TIME'
    canal_event_type_name = 'MYSQL_DB_EVENT_TYPE'
    canal_before_column_suffix = '_before'
    canal_after_column_suffix = '_after'
    canal_update_suffix = '_updated'

    canal_table_filter = ".*\\..*"

    canal_host = 'localhost'
    canal_port = 11111
    canal_username = None
    canal_destination = None
    canal_password = None
    canal_client_id = None

    def add_arguments_for_canal_suffix(self, parser):
        parser.add_argument('--canal_execute_time_name', action='store', type=str, help='mysql execute time name',
                            default=None)
        parser.add_argument('--canal_event_type_name', action='store', type=str, help='mysql event type name',
                            default=None)
        parser.add_argument('--canal_before-column-suffix', action='store', type=str, help='kafka before column suffix',
                            default=None)
        parser.add_argument('--canal_after_column_suffix', action='store', type=str, help='kafka after column suffix',
                            default=None)
        parser.add_argument('--canal_update_suffix', action='store', type=str, help='kafka update column suffix',
                            default=None)

    def add_arguments(self, parser):
        parser.add_argument('--bootstrap_servers', action='store', help='kafka bootstrap servers split by ,',
                            default=None)
        parser.add_argument('--canal_host', action='store', help='canal server host you can set in .env file',
                            default=None)
        parser.add_argument('--canal_port', action='store', type=int, help='canal server port you can set in .env file',
                            default=None)
        parser.add_argument('--canal_destination', action='store', type=str,
                            help='canal server destination you can set in .env file',
                            default=None)

        parser.add_argument('--canal_username', action='store', type=str, help='canal instance username', default=None)
        parser.add_argument('--canal_password', action='store', type=str,
                            help='canal instance password you can set in .env file ', default=None)
        parser.add_argument('--canal_client_id', action='store', type=str,
                            help='canal instance client-id you can set in .env file ', default=str(randint(1, 1000)))
        parser.add_argument('--canal_table_filter', action='store', type=str,
                            help='canal instance table filter like ".*\\\\..*" you can set in .env file ', default=None)
        self.add_arguments_for_canal_suffix(parser)

    def init_global_params(self, options: dict):
        for k, v in options.items():
            if not hasattr(self, k) or k.startswith('_'):
                continue
            if v is None:
                # overwrite by default
                env_value = os.environ.get(k)
                if env_value is None:
                    default_value = getattr(self, k)
                    if default_value is None:
                        print('please set ', k, ' in your env or by command line')
                        exit(1)
                    setattr(self, k, default_value)
                else:
                    if k == 'canal_port':
                        env_value = int(env_value)
                    setattr(self, k, env_value)
            else:
                setattr(self, k, v)

    from canal.protocol.EntryProtocol_pb2 import RowData, EventType, Column

    SUPPORT_TYPE = {
        EventType.INSERT: "kafka_dc",
        EventType.DELETE: "kafka_dc",
        EventType.UPDATE: "kafka_u",
    }

    @classmethod
    def _convert_type(cls, v: Column) -> any:
        val = v.value
        typ = v.sqlType
        if val is None or len(val) == 0:
            return None
        # type id from java.sql.Types.java
        if typ in {-7, -6, 5, 4, -5, 16}:  # int bool 16 also parse int
            return int(val)
        elif typ in {6, 7, 8, 2, 3}:  # float
            return float(val)
        elif typ == 91:  # DATE 91
            return val
        elif typ == 92 and val is not None:  # TIME 92
            return val + "Z"
        elif typ == 93 and val is not None:  # TIMESTAMP 93
            return val.replace(' ', 'T') + 'Z'
        return val

    def _generate_by_columns(self, columns: list, name_suffix: Optional[str] = None) -> dict:
        return {(x.name if name_suffix is None else x.name + name_suffix): self._convert_type(x) for x in columns}

    def _convert_after_column(self, value: any, updated: bool) -> any:
        if updated:
            return self._convert_type(value)
        return None

    def _generate_after_columns(self, columns: list):
        return {(x.name + self.canal_after_column_suffix): self._convert_after_column(x, x.updated) for x in
                columns if x.updated}

    def _generate_notice(self, event_type: int, row: RowData, execute_time: str) -> bytes:
        from canal.protocol.EntryProtocol_pb2 import EventType
        _cv = self._convert_type
        execute_time_name = self.canal_execute_time_name
        event_type_name = self.canal_event_type_name
        if event_type == EventType.DELETE:
            data = self._generate_by_columns(row.beforeColumns)
            data[event_type_name] = event_type
        elif event_type == EventType.INSERT:
            data = self._generate_by_columns(row.afterColumns)
            data[event_type_name] = event_type
        elif event_type == EventType.UPDATE:
            update_suffix = self.canal_update_suffix
            data = {x.name + update_suffix: x.updated for x in row.afterColumns}
            before_data = self._generate_by_columns(row.beforeColumns, self.canal_before_column_suffix)
            after_data = self._generate_after_columns(row.afterColumns)
            data.update(before_data)
            data.update(after_data)
        else:
            raise Exception("Not support Now")
        assert execute_time not in data
        data[execute_time_name] = execute_time
        return json.dumps(data).encode('utf-8')

    def _generate_topic_name(self, database: str, table: str, event_type: int):
        typ = self.SUPPORT_TYPE[event_type]
        return f"{database}_{table}_{typ}"

    @classmethod
    def _convert_utc_time(cls, timestamp: int) -> str:
        date = datetime.fromtimestamp(timestamp / 1000)
        # three way for parse timestamp
        # s = date.strftime("%FT%H:%M:%S.%f")[:-3] + "Z"
        # s = date.isoformat('T', timespec='milliseconds') + 'Z'
        s = date.strftime("%FT%H:%M:%S.%f") + "Z"
        return s

    def run_forever(self, client, producer):

        from canal.protocol import EntryProtocol_pb2
        from canal.protocol.EntryProtocol_pb2 import EntryType
        print(datetime.now(), " start running")
        topics = defaultdict(int)
        sleep_times, send_times = 0, 0
        last_execute_time, add_million_seconds = -1, 0
        while True:
            message = client.get(100)
            entries = message['entries']
            for entry in entries:
                entry_type = entry.entryType
                if entry_type in (EntryType.TRANSACTIONBEGIN, EntryType.TRANSACTIONEND):
                    continue
                row_change = EntryProtocol_pb2.RowChange()
                row_change.MergeFromString(entry.storeValue)
                header = entry.header
                database = header.schemaName
                table = header.tableName
                event_type = header.eventType

                # try add million second when meet same execute time
                fix_execute_time = header.executeTime
                if last_execute_time == header.executeTime:
                    if add_million_seconds < 999:
                        add_million_seconds += 1
                        fix_execute_time = header.executeTime + add_million_seconds
                    else:
                        print('over 1000 event in one seconds')
                        fix_execute_time = header.executeTime + add_million_seconds
                else:
                    last_execute_time = header.executeTime
                    add_million_seconds = 0

                row_time = self._convert_utc_time(fix_execute_time)

                logging.debug(' '.join(str(x) for x in
                                       [row_time, fix_execute_time, header.executeTime, header.logfileOffset,
                                        add_million_seconds]))
                for row in row_change.rowDatas:
                    msg = self._generate_notice(event_type, row, row_time)
                    topic_name = self._generate_topic_name(database, table, event_type)
                    if topic_name not in topics:
                        print(topic_name, 'get')
                    topics[topic_name] += 1
                    producer.send(topic_name, value=msg)
                    send_times += 1
            if not entries:
                time.sleep(WAIT_TIMES)
                if sleep_times % PRINT_EACH == 0:
                    for t, n in topics.items():
                        print(' Topic: ', t, ' send ', n)
                    topics = defaultdict(int)
                    logger.debug("{} wait for in {} already send {}".format(datetime.now(), sleep_times, send_times))
                sleep_times += 1

    def handle(self, *args, **options):
        self.init_global_params(options)
        from kafka import KafkaProducer

        from canal.client import Client
        try:
            producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers.split(','))

            client = Client()
            client.connect(host=self.canal_host, port=self.canal_port)
            client.check_valid(username=self.canal_username.encode(), password=self.canal_password.encode())
            client.subscribe(client_id=self.canal_client_id.encode(), destination=self.canal_destination.encode(),
                             filter=self.canal_table_filter.encode())
            self.run_forever(client, producer)
        finally:
            client.disconnect()
            producer.close()



# -*- coding: utf-8 -*-
import os
import re
import subprocess
import tempfile
import yaml
from typing import Optional
from jinja2 import Template
from terminado.management import NamedTermManager
from fsqlfly.settings import FSQLFLY_UPLOAD_DIR, FSQLFLY_FLINK_BIN, logger
from fsqlfly.models import Transform, Namespace, Resource, Functions
from fsqlfly import settings


def _create_config_from_resource(resource: Resource, **kwargs) -> dict:
    data = yaml.load(handle_template(resource.yaml, **kwargs), yaml.FullLoader)
    data['name'] = resource.name
    return data


def _get_functions() -> list:
    res = []
    for f in Functions.select().where(Functions.is_deleted == False).objects():
        fc = {
            'name': f.name,
            'from': f.function_from,
            'class': f.class_name
        }
        if f.constructor_config.strip():
            fc['constructor'] = yaml.load(f.constructor_config, yaml.FullLoader)
        res.append(fc)
    return res


def _get_jar() -> str:
    res = []
    for f in Functions.select().where(Functions.is_deleted == False).objects():
        r_p = os.path.join(FSQLFLY_UPLOAD_DIR, f.resource.real_path[1:])
        res.append(r_p)

    out = []
    for x in set(res):
        out.extend(['-j', x])
    return out


def _create_config(require: str, config: Optional[str], **kwargs) -> str:
    tables = []
    if require.strip():
        for x in require.split(','):
            if '.' in x:
                space, name = x.split('.', 1)
            else:
                space, name = None, x

            if space:
                space = Namespace.select().where(Namespace.name == space).first()

            resource = Resource.select().where(Namespace.name == name & Resource.namespace == space).first()
            resource_only_name = Resource.select().where(Resource.name == name).first()
            if resource is None and resource_only_name is not None:
                resource = resource_only_name

            data = _create_config_from_resource(resource, **kwargs)
            tables.append(data)
    base_config = yaml.load(handle_template(config, **kwargs), yaml.FullLoader) if config else dict()
    if base_config is None:
        base_config = dict()
    if base_config.get('tables'):
        base_config['tables'].extend(tables)
    else:
        base_config['tables'] = tables
    base_config['functions'] = _get_functions()
    return yaml.dump(base_config)


def _clean(s: str) -> str:
    return re.sub(r'\[\d+(;\d+)?m?', '', s)


def handle_template(text: str, **kwargs) -> str:
    return Template(text).render(**kwargs)


def run_transform(transform: Transform, **kwargs) -> (bool, str):
    _, yaml_f = tempfile.mkstemp(suffix='.yaml')
    _, sql_f = tempfile.mkstemp(suffix='.sql')

    yaml_conf = _create_config(transform.require, transform.yaml, **kwargs)
    sql = handle_template(transform.sql, **kwargs)
    print(yaml_conf, file=open(yaml_f, 'w'))
    print(sql, file=open(sql_f, 'w'))
    print('q\nexit;', file=open(sql_f, 'a+'))
    pt = '_' + kwargs['pt'] if 'pt' in kwargs else ''
    run_commands = [FSQLFLY_FLINK_BIN, 'embedded',
                    '-s', '{}_{}.{}-'.format(transform.id, transform.name, pt),
                    '--environment', yaml_f,
                    *_get_jar(),
                    '<', sql_f]
    print(' '.join(run_commands))
    try:
        out = subprocess.check_output(' '.join(run_commands), shell=True, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as error:
        return False, "sql:\n {} \n\noutput: \n{} \n\n error: {}\n\nyaml: \n{}".format(transform.sql,
                                                                                       error.stdout.decode(),
                                                                                       error.stderr.decode(),
                                                                                       yaml_conf)
    except Exception as e:
        print(e)
        return False, str(e)

    out_w = _clean(out.decode())

    os.remove(yaml_f)
    os.remove(sql_f)

    print(out_w, file=open('out.shell.txt', 'w'))
    print(out_w)
    return True, out_w


def run_debug_transform(data: dict, manager: NamedTermManager) -> (str, str):
    _, yaml_f = tempfile.mkstemp(suffix='.yaml')
    yaml_conf = _create_config(data.get('require', ''), data.get('yaml', ''))
    print(yaml_conf, file=open(yaml_f, 'w'))
    real_sql = handle_template(data.get('sql', ''))
    name = manager._next_available_name()
    run_commands = [FSQLFLY_FLINK_BIN, 'embedded',
                    '-s', '{}{}.'.format(settings.TEMP_TERMINAL_HEAD, str(name)),
                    '--environment', yaml_f,
                    *_get_jar()]
    logger.debug('running commands is : {}'.format(' '.join(run_commands)))
    term = manager.new_terminal(shell_command=run_commands)

    logger.debug('sql :{}'.format(real_sql))
    term.ptyproc.write(real_sql)
    term.term_name = name
    setattr(term, settings.TERMINAL_OPEN_NAME, True)
    term.run_command = ' '.join(run_commands)
    manager.terminals[name] = term
    return name



import os
import sys
import hashlib
import logging
import logging
import logzero
from logzero import logger
from os.path import join
from dotenv import load_dotenv
from pathlib import Path
from peewee import MySQLDatabase, SqliteDatabase, PostgresqlDatabase
from playhouse.cockroachdb import CockroachDatabase
from playhouse.db_url import connect
from terminado.management import NamedTermManager


def generate_cookie_secret(s: str, typ: str = '___cookie_secret') -> str:
    md5 = hashlib.md5()
    md5.update(f'{s}{typ}'.encode())
    return md5.hexdigest()


BASE_DIR = os.path.expanduser('~')
ENV_FILE_PATH_FROM_LOCAL_ENV = os.environ.get('FSQLFLY')
ENV_FILE_PATH = join(BASE_DIR, '.fsqlfly')
if ENV_FILE_PATH_FROM_LOCAL_ENV and os.path.exists(ENV_FILE_PATH_FROM_LOCAL_ENV):
    logger.debug('Change Env File to {}'.format(ENV_FILE_PATH_FROM_LOCAL_ENV))
    ENV_FILE_PATH = ENV_FILE_PATH_FROM_LOCAL_ENV
if os.path.exists(ENV_FILE_PATH) and os.path.isfile(ENV_FILE_PATH):
    logging.debug("Load Env File From {} ".format(ENV_FILE_PATH))
    load_dotenv(dotenv_path=Path(ENV_FILE_PATH))
else:
    logging.debug("Not Found Valid Env File ".format(ENV_FILE_PATH))

ENV = os.environ.get
FSQLFLY_DEBUG = ENV('FSQLFLY_DEBUG') is not None
if FSQLFLY_DEBUG:
    logzero.loglevel(logging.DEBUG)

FSQLFLY_PASSWORD = ENV('FSQLFLY_PASSWORD', 'password')
logger.debug('FSQLFLY_PASSWORD is {}'.format(FSQLFLY_PASSWORD))
FSQLFLY_TOKEN = generate_cookie_secret(FSQLFLY_PASSWORD, '')
FSQLFLY_TOKEN_BYTE = FSQLFLY_TOKEN.encode()
logger.debug('FSQLFLY_TOKEN_BYTE is {}'.format(FSQLFLY_TOKEN_BYTE))

FSQLFLY_COOKIE_SECRET = generate_cookie_secret(FSQLFLY_PASSWORD)

FSQLFLY_DB_URL = ENV('FSQLFLY_DB_URL')
FSQLFLY_DB_TYPE = ENV('FSQLFLY_DB_TYPE', 'sqlite')
FSQLFLY_DB_FILE = ENV('FSQLFLY_DB_FILE', 'db.sqlite3')
FSQLFLY_DATABASE = ENV('FSQLFLY_DATABASE', 'test')
FSQLFLY_DB_PASSWORD = ENV('FSQLFLY_DB_PASSWORD', 'xxx')
FSQLFLY_DB_USER = ENV('FSQLFLY_DB_USER', 'root')
FSQLFLY_DB_HOST = ENV('FSQLFLY_DB_HOST', 'localhost')
FSQLFLY_DB_PORT = int(ENV('FSQLFLY_DB_PORT', '3306'))

if FSQLFLY_DB_URL is not None:
    DATABASE = connect(FSQLFLY_DB_URL)
    logger.debug('DATABASE is connect by url : {}'.format(FSQLFLY_DB_URL))
elif FSQLFLY_DB_TYPE == 'sqlite':
    DATABASE = SqliteDatabase(FSQLFLY_DB_FILE)
    logger.debug('DATABASE is connect by sqlite: {}'.format(FSQLFLY_DB_URL))
else:
    assert FSQLFLY_DB_TYPE in (
        'mysql', 'postgresql', 'cockroach'), 'env FSQLFLY_DB_TYPE must in (sqlite|mysql|postgresql|cockroach)'
    db_classes = {
        'mysql': MySQLDatabase,
        'postgresql': PostgresqlDatabase,
        'cockroach': CockroachDatabase
    }
    db_class = db_classes[FSQLFLY_DB_TYPE]
    logger.debug('DATABASE is connect by {}: {}:{}@{}:{}/{}'.format(FSQLFLY_DB_TYPE,
                                                                    FSQLFLY_DB_USER, FSQLFLY_DB_PASSWORD,
                                                                    FSQLFLY_DB_HOST, FSQLFLY_DB_PORT,
                                                                    FSQLFLY_DATABASE))

    DATABASE = db_class(database=FSQLFLY_DATABASE,
                        password=FSQLFLY_DB_PASSWORD, user=FSQLFLY_DB_USER, host=FSQLFLY_DB_HOST, port=FSQLFLY_DB_PORT)

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
FSQLFLY_UPLOAD_DIR = ENV('FSQLFLY_UPLOAD_DIR', join(os.path.expanduser('~'), '.fsqlfly_upload'))
if not os.path.exists(FSQLFLY_UPLOAD_DIR):
    logger.debug('Create Upload Base Dir {}'.format(FSQLFLY_UPLOAD_DIR))
    os.makedirs(FSQLFLY_UPLOAD_DIR)
UPLOAD_ROOT_DIR = join(FSQLFLY_UPLOAD_DIR, 'upload')

FSQLFLY_STATIC_ROOT = ENV('FSQLFLY_STATIC_ROOT', join(ROOT_DIR, 'static'))
FSQLFLY_FINK_HOST = ENV('FSQLFLY_FINK_HOST', 'http://localhost:8081')
FSQLFLY_JOB_DAEMON_FREQUENCY = int(ENV('FSQLFLY_JOB_DAEMON_FREQUENCY', '30'))
FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY = int(ENV('FSQLFLY_JOB_DAEMON_MAX_TRY_ONE_DAY', '3'))

assert os.path.exists(FSQLFLY_STATIC_ROOT), "FSQLFLY_STATIC_ROOT ({}) not set correct".format(FSQLFLY_STATIC_ROOT)
INDEX_HTML_PATH = join(FSQLFLY_STATIC_ROOT, 'index.html')
assert os.path.exists(INDEX_HTML_PATH), "FSQLFLY_STATIC_ROOT ({}) can't find index.html".format(FSQLFLY_STATIC_ROOT)
INDEX_HTML = open(INDEX_HTML_PATH, 'rb').read()

FSQLFLY_FLINK_BIN_DIR = ENV('FSQLFLY_FLINK_BIN_DIR', '/opt/flink/bin')
assert os.path.exists(FSQLFLY_FLINK_BIN_DIR), "FSQLFLY_FLINK_BIN_DIR ({}) not exists, please confirm".format(
    FSQLFLY_FLINK_BIN_DIR)
FSQLFLY_FLINK_BIN = join(FSQLFLY_FLINK_BIN_DIR, 'sql-client.sh')

FSQLFLY_FLINK_MAX_TERMINAL = int(ENV('FSQLFLY_FLINK_MAX_TERMINAL', '100'))
FSQLFLY_WEB_PORT = int(ENV('FSQLFLY_WEB_PORT', '8082'))

TERMINAL_MANAGER = NamedTermManager(
    shell_command=[FSQLFLY_FLINK_BIN, 'embedded'],
    max_terminals=FSQLFLY_FLINK_MAX_TERMINAL)

TERMINAL_OPEN_NAME = '__NEED_OPEN_'
FSQLFLY_JOB_LOG_DIR = ENV('FSQLFLY_JOB_LOG_DIR', '/tmp/fsqlfly_job_log')
FSQLFLY_JOB_LOG_FILE = join(FSQLFLY_JOB_LOG_DIR, 'job_damon.log')
os.makedirs(FSQLFLY_JOB_LOG_DIR, exist_ok=True)

TEMP_TERMINAL_HEAD = '0__TEMPORARY__'




# -*- coding:utf-8 -*-
import json
import logging
from typing import Any
from abc import ABC
from datetime import datetime, date
import tornado
import tornado.web
from fsqlfly import settings


class RespCode:
    Success = 200
    ServerError = 500
    NeedLogin = 503
    LoginFail = 501
    APIFail = 502
    InvalidHttpMethod = 405


def json_obj_hook(o):
    if isinstance(o, (date, datetime)):
        return str(o)


class BaseHandler(tornado.web.RequestHandler):
    """Request handler where requests and responses speak JSON."""

    @property
    def terminal_manager(self):
        return self.settings['terminal_manager']

    def get_current_user(self) -> Any:
        cookie = self.get_cookie('user')
        if cookie:
            return cookie
        token = self.request.arguments.get('token')
        if token is None:
            token = self.request.headers.get('token')
        logging.debug("Try Login By token {}".format(token))

        if token is not None and len(token) == 1 and token[0] == settings.FSQLFLY_TOKEN_BYTE:
            logging.debug("Login By token {}".format(token))
            return token
        return None

    def get_login_url(self) -> str:
        return '/login'

    def set_login_status(self):
        self.set_secure_cookie('user', 'admin')

    def set_logout_status(self):
        self.clear_cookie('user')

    @property
    def json_body(self) -> dict:
        _name = '_save_json_data'
        if not hasattr(self, _name):
            try:
                json_data = json.loads(self.request.body)
                setattr(self, _name, json_data)
            except ValueError:
                setattr(self, _name, {})

        return getattr(self, _name)

    def set_default_headers(self):
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    code_msg = {
        RespCode.NeedLogin: 'Need Login',
        RespCode.ServerError: 'Web Server Error',
        RespCode.LoginFail: 'Wrong Password or Token',
        RespCode.InvalidHttpMethod: 'Invalid HTTP method.',
        RespCode.APIFail: 'Invalid API Request',
    }

    def write_error(self, status_code, **kwargs):
        kwargs['code'] = status_code
        kwargs['success'] = status_code == 200
        if 'msg' not in kwargs:
            kwargs['msg'] = self.code_msg.get(status_code, 'Unknown error.')
        self.set_status(200)
        self.write_json(kwargs)

    def write_json(self, res: dict):
        output = json.dumps(res, ensure_ascii=False, default=json_obj_hook)
        self.write(output)
        self.finish()



import time
from tornado.ioloop import IOLoop
from tornado import gen
import tormysql
import tornado.web
import asyncio
from peewee_async import Manager, PostgresqlDatabase, MySQLDatabase
from playhouse.db_url import connect
from tornado.web import Application

loop = asyncio.new_event_loop()
user, password, host, db = 'root', 'zhanglun', 'localhost', 'sample'
database = connect(f'mysql://{user}:{password}@{host}:3306/{db}')
connection_pool = tormysql.helpers.ConnectionPool(
    max_connections=20,  # max open connections
    idle_seconds=7200,  # conntion idle timeout time, 0 is not timeout
    wait_connection_timeout=3,  # wait connection timeout
    host=host,
    user=user,
    passwd=password,
    db=db,
    charset="utf8"
)

from peewee import Model, CharField, DateField
from datetime import datetime
from playhouse.shortcuts import model_to_dict


class BaseModel(Model):
    class Meta:
        database = database


class Person(BaseModel):
    name = CharField()
    create_at = DateField(default=datetime.now)


database.drop_tables([Person], safe=True)
database.create_tables([Person], safe=True)
for i in range(1000):
    Person.create(name=f'test {i}')

from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor


def get_all_person():
    return str(list(Person.select().objects()))


class AsyncHandler(tornado.web.RequestHandler):
    executor = ThreadPoolExecutor()

    async def get(self):
        person = await self.get_person()
        self.write(person)

    @run_on_executor
    def get_person(self):
        return get_all_person()


class SyncHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(get_all_person())


class PoolHandler(tornado.web.RequestHandler):

    async def get(self):
        cursor = await connection_pool.execute("select * from person")
        data = cursor.fetchall()
        self.write(str(data))


def make_app():
    return tornado.web.Application([
        (r"/sync", SyncHandler),
        (r"/async", AsyncHandler),
        (r"/pool_async", PoolHandler),
    ],
        debug=True)


app = make_app()
app.listen(8889)

if __name__ == "__main__":
    tornado.ioloop.IOLoop.current().start()



# -*- coding:utf-8 -*-
import os
import magic
from tempfile import mkstemp
import tornado.web
from tornado.web import authenticated
from fsqlfly.base_handle import BaseHandler, RespCode
from fsqlfly.utils.response import create_response
from fsqlfly import settings
from fsqlfly.settings import UPLOAD_ROOT_DIR, FSQLFLY_JOB_LOG_FILE

is_login = False
user = dict(code=200, name='flink', status='ok', currentAuthority='admin', type='password',
            avatar='https://gw.alipayobjects.com/zos/antfincdn/XAosXuNZyF/BiazfanxmamNRoxxVxka.png')

support_upload = ['logo', 'file']
upload_dirs = {x: os.path.join(UPLOAD_ROOT_DIR, x) for x in support_upload}
_ = [os.makedirs(x, exist_ok=True) for x in upload_dirs.values()]

FileMagic = magic.Magic(mime=True, uncompress=True)


class UploadHandler(BaseHandler):
    @authenticated
    def get(self, path: str):
        if not any(map(lambda x: path.startswith(x), support_upload)):
            raise tornado.web.HTTPError(status_code=404)
        full_path = os.path.join(UPLOAD_ROOT_DIR, path)
        if not os.path.exists(full_path):
            self.write_error(RespCode.APIFail)
        mime = FileMagic.from_file(full_path)
        self.set_header('content-type', mime)
        self.write(open(full_path, "rb").read())
        self.finish()

    @authenticated
    def post(self):
        files = self.request.files
        key = list(files.keys())[0]
        if key not in support_upload:
            self.write_error(RespCode.APIFail)
        upload_file = files[key][0]

        _, tem_f = mkstemp(suffix=upload_file.filename, dir=upload_dirs[key])
        with open(tem_f, 'wb+') as out:
            out.write(upload_file.body)
        real_path = '/upload/' + key + '/' + os.path.basename(tem_f)
        return self.write_json(create_response(data={"realPath": real_path}))


class LogHandler(BaseHandler):
    @authenticated
    def get(self):
        if not os.path.exists(FSQLFLY_JOB_LOG_FILE):
            raise tornado.web.HTTPError(status_code=404)
        self.set_header('content-type', 'text')
        self.write(open(FSQLFLY_JOB_LOG_FILE, "rb").read())
        self.finish()


class BlankHandler(BaseHandler):
    def get(self):
        self.set_header('content-type', 'text')
        self.write('...')
        self.finish()


default_handlers = [
    (r'/api/upload', UploadHandler),
    (r'/api/log', LogHandler),
    (r'/api/blank', BlankHandler),
    (r'/upload/(?P<path>.+)', UploadHandler),
]



import unittest
import importlib
from fsqlfly import settings
import fsqlfly.models
from peewee import SqliteDatabase, IntegrityError
from unittest.mock import patch


def gen_mock(f):
    def _gen(*args, **kwargs):
        with patch('fsqlfly.settings.DATABASE', SqliteDatabase(':memory:')):
            return f(*args, **kwargs)

    return _gen


class BaseTestCase(unittest.TestCase):
    @gen_mock
    def setUp(self) -> None:
        fsqlfly.models.create_all_tables()

    @gen_mock
    def tearDown(self) -> None:
        fsqlfly.models.delete_all_tables(force=True)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
import json
import tornado.web
from terminado import TermSocket
from tornado.web import authenticated
from logzero import logger
from fsqlfly import settings
from fsqlfly.base_handle import BaseHandler, RespCode
from fsqlfly.models import Namespace, FileResource, Functions, Transform, Resource
from fsqlfly.utils.strings import dict2camel, dict2underline
from fsqlfly.utils.response import create_response

_MODELS = {
    'functions': Functions,
    'namespace': Namespace,
    'resource': Resource,
    'transform': Transform,
    'file': FileResource,
}


class APICounter(BaseHandler):
    @authenticated
    def get(self):
        data = {
            "namespaceNum": Namespace.select().where(Namespace.is_deleted == False).count(),
            "resourceNum": Resource.select().where(Resource.is_deleted == False).count(),
            "functionNum": Functions.select().where(Functions.is_deleted == False).count(),
            "fileNum": FileResource.select().where(FileResource.is_deleted == False).count(),
            "transformNum": Transform.select().where(Transform.is_deleted == False).count(),
            "code": 200, "success": True
        }
        return self.write_json(data)


class CRHandler(BaseHandler):
    @authenticated
    def get(self, model: str):
        if model not in _MODELS:
            return self.write_error(RespCode.APIFail)

        db = _MODELS[model]
        data = [x.to_dict(to_camel=True) for x in db.select().where(db.is_deleted == False).objects()]
        return self.write_json(create_response(data))

    @authenticated
    def post(self, model: str):
        if model not in _MODELS:
            return self.write_error(RespCode.APIFail)
        data = dict2underline(self.json_body)
        same_obj = _MODELS[model].select().where(_MODELS[model].name == data['name']).first()
        if same_obj:
            logger.debug('try create a same name obj {} @{}'.format(same_obj.name, model))
            if same_obj.is_deleted:
                logger.debug('try delete a same name obj {} @{}'.format(same_obj.name, model))
                same_obj.delete_instance()
            else:
                return self.write_error(RespCode.APIFail, msg='Same name {} in {}'.format(same_obj.name, model))
        _obj = _MODELS[model].create(
            **{k: v for k, v in data.items() if not (k.endswith('_id') and data[k] == 0)})
        self.write_json(create_response(data=_obj.to_dict()))


class UDHandler(BaseHandler):
    @authenticated
    def post(self, model: str, pk: int):
        if model not in _MODELS:
            return self.write_error(RespCode.APIFail)
        obj = _MODELS[model].select().where(_MODELS[model].id == pk).get()
        data = self.json_body
        if 'name' in data:
            if data['name'] != obj.name:
                return self.write_error(RespCode.APIFail, msg='创建之后禁止修改名字')

        for k, v in dict2underline(data).items():
            if k not in ('update_at', 'create_at', 'is_deleted', 'id') and not (k.endswith('_id') and v == 0):
                setattr(obj, k, v)
        obj.save()
        self.write_json(create_response(data=obj.to_dict()))

    @authenticated
    def delete(self, model: str, pk: int):
        if model not in _MODELS:
            return self.write_error(RespCode.APIFail)
        obj = _MODELS[model].select().where(_MODELS[model].id == pk).get()
        if obj is None:
            return self.write_error(RespCode.APIFail, msg='Object id {} not exist {}'.format(pk, model))
        obj.is_deleted = True
        obj.save()
        self.write_json(create_response(data=obj.to_dict()))


default_handlers = [
    (r'/api/count', APICounter),
    (r'/api/(?P<model>\w+)', CRHandler),
    (r'/api/(?P<model>\w+)/(?P<pk>\d+)', UDHandler),
]



# -*- coding: utf-8 -*-
from typing import Optional, Union

from fsqlfly.utils.strings import dict2camel


def create_response(data: Union[Optional[dict], list] = None, code: int = 0, msg: str = '') -> dict:
    return dict2camel({'data': data, 'code': code, 'success': code == 0, 'msg': msg})



import asyncio
import time
import tornado.web
from terminado import TermSocket
from tornado.web import authenticated
from fsqlfly import settings
from fsqlfly.base_handle import BaseHandler, RespCode
from fsqlfly.workflow import run_debug_transform
from fsqlfly.utils.response import create_response
from fsqlfly.models import Transform


class TerminalHandler(BaseHandler):
    @authenticated
    def get(self):
        tm = self.terminal_manager
        terms = [{'name': name, 'id': name} for name in tm.terminals]
        self.write_json(create_response(data=terms))


class TerminalNewHandler(BaseHandler):
    @authenticated
    def post(self, mode: str, pk: int):
        if mode == 'debug':
            term = run_debug_transform(self.json_body, self.terminal_manager)
            self.write_json(create_response({"url": '/terminal/{}'.format(term)}))

        else:
            self.redirect('/api/job/{}/{}'.format(mode, pk))


class TerminalStopHandler(BaseHandler):
    @authenticated
    async def post(self, name: str):
        tm = self.terminal_manager
        if name in tm.terminals:
            await tm.terminate(name, force=True)
            self.write_json(create_response())
        else:
            raise tornado.web.HTTPError(404, "Terminal not found: %r" % name)


class MyTermSocket(TermSocket):
    def open(self, url_component=None):
        super(MyTermSocket, self).open(url_component)
        terminal = self.term_manager.get_terminal(url_component)
        open_name = settings.TERMINAL_OPEN_NAME
        if hasattr(terminal, open_name) and getattr(terminal, open_name):
            setattr(terminal, open_name, False)
            self.term_manager.start_reading(terminal)



default_handlers = [
    (r'/api/terminal', TerminalHandler),
    (r"/_websocket/(\w+)", MyTermSocket, {'term_manager': settings.TERMINAL_MANAGER}),
    (r'/api/terminal/stop/(?P<name>\d+)', TerminalStopHandler),
    (r'/api/transform/(\w+)/(\d+)', TerminalNewHandler),
]



### Test Result
### In 4c-2.9g computer get 1000 result
### async  cost  1.8666772842407227  s  0.03111128807067871  min  speed  535.7112385962062  c/s
### sync  cost  1.8722515106201172  s  0.03120419184366862  min  speed  534.1162735495859  c/s
### pool_async  cost  10.745956897735596  s  0.1790992816289266  min  speed  93.05825525977323  c/s
###

import time
import unittest
from unittest import TestCase
import requests
from concurrent.futures import ThreadPoolExecutor


class TestSpeed(TestCase):
    def run_once(self, path):
        res = requests.get(f'http://localhost:8889/{path}')
        # print(res.text)

    def run_path(self, path):
        start = time.time()
        num = 1000
        pool = ThreadPoolExecutor()
        for i in range(num):
            pool.submit(self.run_once, path)
        pool.shutdown()
        cost = time.time() - start

        print(path, ' cost ', cost, ' s ', cost / 60, ' min', ' speed ', num / cost, ' c/s')
        return cost

    def test_main(self):
        # async_cost = self.run_path('async')
        # sync_cost = self.run_path('sync')
        pool_async_cost = self.run_path('pool_async')
        # assert async_cost < pool_async_cost < sync_cost


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

from fsqlfly.base_handle import BaseHandler, RespCode
from fsqlfly import settings

is_login = False
user = dict(code=200, name='admin', status='ok', currentAuthority='admin', type='password',
            token=settings.FSQLFLY_TOKEN, success=True)


class LoginHandler(BaseHandler):
    def get(self):
        if self.current_user is None:
            self.write_error(500)
        else:
            self.write_json(user)

    def post(self):
        print(self.request.arguments)
        arg = self.json_body
        if (arg.get('type') == 'account' and arg.get('password') == settings.FSQLFLY_PASSWORD) or (
                arg.get('type') == 'token' and arg.get('token') == settings.FSQLFLY_TOKEN
        ):
            self.set_login_status()
            self.write_json(user)
        else:
            self.write_error(RespCode.LoginFail)


class LoginOutHandler(BaseHandler):
    def get(self):
        if self.current_user:
            self.set_logout_status()
            self.write_json(user)


default_handlers = [
    (r'/api/login', LoginHandler),
    (r'/api/logout', LoginOutHandler),
]



# -*- coding: utf-8 -*-
import re
import yaml
from typing import Optional
from collections import namedtuple
from io import StringIO
from typing import Callable, List, Tuple

__CAMEL_PATTERN = re.compile("(?<=[a-z0-9])(_[a-z])")


def str2camel(v: str) -> str:
    return __CAMEL_PATTERN.sub(lambda x: x.group()[1].upper(), v)


__UNDERLINE_PATTERN = re.compile("(?<=[a-z0-9])([A-Z])")


def str2underline(v: str) -> str:
    return v[0].lower() + __UNDERLINE_PATTERN.sub(lambda x: '_' + x.group().lower(), v[1:])


def _dict2_(value: dict, method: Callable[[str], str]) -> dict:
    return {method(k): _dict2_(v, method) if isinstance(v, dict) else (
        [_dict2_(xx, method) for xx in v] if isinstance(v, list) else v) for k, v in
            value.items()}


def dict2camel(v: dict) -> dict:
    return _dict2_(v, str2camel)


def dict2underline(v: dict) -> dict:
    return _dict2_(v, str2underline)


def _load_yaml(source: str) -> dict:
    d = yaml.load(StringIO(source), Loader=yaml.SafeLoader)
    return d


def check_yaml(source: str) -> bool:
    try:
        _load_yaml(source)
    except Exception as e:
        print(e)
        return False
    else:
        return True


def get_schema(source: str) -> dict:
    data = _load_yaml(source)
    if 'schema' not in data:
        raise SystemError("Schema Not In Yaml file")
    return data['schema']


def get_sink_config(source: str) -> dict:
    return _load_yaml(source)


def generate_yaml(source: List[Tuple[str, str]], is_sink: bool = False) -> str:
    schemas = []
    for t, d in source:
        if d == 'timestamp' and not is_sink:
            schemas.append({'name': 'rowTime', 'type': 'TIMESTAMP', 'rowtime': {'timestamps':
                                                                                    {'type': 'from-field',
                                                                                     'from': d},
                                                                                'watermarks': {
                                                                                    'type': 'periodic-bounded',
                                                                                    'delay': '6000'
                                                                                }}})
        else:
            schemas.append({'name': d, 'type': t})
    data = {
        'format': {
            'property-version': 1,
            'type': 'json',
            'schema': 'ROW<{}>'.format(', '.join(
                (x['name'] if 'rowtime' not in x else x['rowtime']['timestamps']['from']) + ' ' + x['type'] for x in
                schemas))
        },
        'schema': schemas
    }
    out = yaml.dump(data)
    return out


SqlProps = namedtuple('SqlProps', ['name', 'value'])

SQL_COMMENT_PATTERN = re.compile(r'/\*(.*?)(?=\*/)', re.DOTALL)
SQL_COMMENT_PATTERN_CLEAN = re.compile(r'/\*.*?(?=\*/)\*/', re.DOTALL)


def parse_sql(sql: str) -> List[SqlProps]:
    res = []
    for x in SQL_COMMENT_PATTERN.findall(sql):
        if '=' in x:
            k, v = map(str.strip, x.split('=', maxsplit=1))
            if v.startswith("'") or v.startswith('"'):
                v = v[1:-1]
            elif re.findall('\d+\.\d+', v):
                v = float(v)
            elif v.isdigit():
                v = int(v)
            res.append(SqlProps(k, v))
        else:
            res.append(SqlProps(x.strip(), True))
    return res


def clean_sql(sql: str) -> str:
    return SQL_COMMENT_PATTERN_CLEAN.sub('', sql).strip()
